/*********************************************************
	SDRAW.H -- Header file for "simple" Windows DC / GDI
				  encapsulation.

				Includes support for:

				-- Basic GDI drawing--brush/pen/font/ROP2 mode
					manipulation.
				-- Off-screen buffering.
				-- Draws other SWindows objects--SBitmap's,
					SSprite's, SCartoon's.
*********************************************************/

#if !defined (_SDRAW_H)
	#define _SDRAW_H
#include <windows.h>
#include <stdlib.h>
#include "sdrawcor.h"
#include "sbitmap.h"
#include "ssprite.h"
#include "scartoon.h"

#define INCL_DEBUG

// Helper function returns a valid RGB value from COLORS.H
DWORD GetRandomColor(void);

class SDraw
{
private:
	SDrawCore *pCoreModule;
	int bCoreWasCreatedByUs;

	HWND hwndFrame;
	HDC hdcWindow;

	int nDefaultWidth;

	PAINTSTRUCT ps;

	BOOL bGraphicsInitOk;

public:
	// Constructor
	SDraw(HWND hwnd, SDrawCore *pCore = NULL);

	// Destructor
	~SDraw();

	// For non-WM_PAINT event drawing...
	void StartDrawing(int bDuringPaintEvent = FALSE);
	void StopDrawing(int bDuringPaintEvent = FALSE);

	// For WM_PAINT Painting only...
	void StartPainting(void) { StartDrawing(TRUE); }
	void StopPainting(void) { StopDrawing(TRUE); }

	// GDI API encapsulation.
	void SetPixel(int x, int y, DWORD Color);
	void SetPixel(POINT pt, DWORD Color);

	void GetPixel(int x, int y, DWORD& Color);
	void GetPixel(POINT pt, DWORD& Color);
	DWORD GetPixel(int x, int y);

	void Line(int x1, int y1, int x2, int y2, DWORD Color, int nWidth);
	void Line(POINT pt1, POINT pt2, DWORD Color, int nWidth);

	void MoveTo(int x, int y);
	void MoveTo(POINT pt);

	void LineTo(int x, int y, DWORD Color, int nWidth);
	void LineTo(POINT pt, DWORD Color, int nWidth);

	void Rectangle(int x1, int y1, int x2, int y2, DWORD Color, int bBorderOnly = FALSE);
	void Rectangle(POINT pt1, POINT pt2, DWORD Color, int bBorderOnly = FALSE);

	void Ellipse(int x1, int y1, int x2, int y2, DWORD Color, int bBorderOnly = FALSE);
	void Ellipse(POINT pt1, POINT pt2, DWORD Color, int bBorderOnly = FALSE);

	// (Sorry, Arc and Pie support not included in this version.)

	void SetLineWidth(int nWidth = 1);

	void SetPixelCombinationMethod(int nComboType);
	void SetROP2(DWORD dwROP2);

	// Font and text manipulation
	void SetFont(LPSTR lpszFaceName, int nPoints, int bIsBold = FALSE);

	void SetTextForeColor(DWORD FColor);
	void SetTextBackColor(DWORD BColor);

	void DisplayText(int x, int y, LPSTR lpszText, DWORD Color = -1);
	void DisplayText(POINT pt, LPSTR lpszText, DWORD Color = -1);
	void DisplayText(RECT& r, LPSTR lpszText, DWORD Color = -1);

#if defined (INCL_DEBUG)
	void ReportToolUsage(HWND hwnd);
#endif

	void TextOut(int x, int y, LPSTR lpszText, int nLength = -1);
	void DrawText(LPSTR lpszText, RECT& rClip, int nOptions = DT_VCENTER | DT_WORDBREAK);
	void ExtTextOut(int x, int y, UINT uFlags, RECT& rClip, LPSTR lpszText, int nLength = -1);

	// Image/bitmap routines.
	void DrawBitmap(int x, int y, SBitmap& image, DWORD dwROP = SRCCOPY);
	void DrawBitmap(POINT pt, SBitmap& image, DWORD dwROP = SRCCOPY);

	void StretchBitmap(int x, int y, int nNewWidth, int nNewHeight, SBitmap& image, DWORD dwROP = SRCCOPY);
	void StretchBitmap(POINT pt1, POINT pt2, SBitmap& image, DWORD dwROP = SRCCOPY);

	// Sprite routines.
	void DrawSprite(int x, int y, SSprite& sprite);
	void DrawSprite(POINT pt, SSprite& sprite);
	void DrawSprite(SSprite& sprite);
	void DrawSprite(POINT pt, SSprite& sprite, RECT& rXORBmp, RECT& rANDBmp);

	void StretchSprite(int nNewWidth, int nNewHeight, SSprite& sprite);
	void StretchSprite(int x, int y, int nNewWidth, int nNewHeight, SSprite& sprite);
	void StretchSprite(POINT pt1, POINT pt2, SSprite& sprite);

	void DrawCartoon(int x, int y, SCartoon& cartoon);
	void DrawCartoon(POINT pt, SCartoon& cartoon) { DrawCartoon(pt.x, pt.y, cartoon); }
	void DrawCartoon(SCartoon& cartoon) { DrawCartoon(cartoon.x, cartoon.y, cartoon); }

	void StartScreenBuffering(void);
	void StopScreenBuffering(void);

	void CopyScreenToBuffer(int cxCopy = 0, int cyCopy = 0);
	void CopyBufferToScreen(int cxCopy = 0, int cyCopy = 0);

	HDC GetActiveScreen(void) { if(bGraphicsInitOk) return pCoreModule->hdcActive; else return NULL; }

	// 'Forward' these requests to the core module...
	BOOL SelectBrush(HDC hdc, DWORD rgbColor) { if (bGraphicsInitOk) return pCoreModule->SelectBrush(hdc, rgbColor); else return FALSE; };
	BOOL SelectPen(HDC hdc, DWORD rgbColor, int nPenWidth) { if (bGraphicsInitOk) return pCoreModule->SelectPen(hdc, rgbColor, nPenWidth); else return FALSE; }
	BOOL SelectFont(HDC hdc, char *lpszFaceName, int nPoints, int bIsBold) { if (bGraphicsInitOk) return pCoreModule->SelectFont(hdc, lpszFaceName, nPoints, bIsBold); else return FALSE; }

	// Allow other objects to poll graphics state.
	BOOL IsGraphicsOk(void) {  return bGraphicsInitOk; };
};
#endif
