/*********************************************************
	SDRAWCOR.CPP -- Implementation file for "simple" Windows DC /
						 GDI encapsulation -- Core routines to manage
						 off-screen buffering and GDI tool
						 management.
*********************************************************/
#include "sdrawcor.h"

int TOOLDESC::CreateIt(void)
{
	if(hTool)
	{
		// Make sure a tool doesn't already exist...
		DeleteObject(hTool);
		hTool = NULL;
	}

	if(TypeCode == Brush)
		hTool = CreateSolidBrush(rgbColor);
	else if(TypeCode == Pen)
		hTool = CreatePen(PS_SOLID, nPenWidth, rgbColor);
	else
	{
		MessageBox(NULL, "Warning: Unexpected Tool Type in TOOLDESC::CreateIt()", "Error", MB_ICONSTOP);
	}

	if(hTool)
		return TRUE;
	else
		return FALSE;
}


void TOOLDESC::DeleteIt(void)
{
	if(hTool)
		// Make sure a tool doesn't already exist...
		DeleteObject(hTool);

	hTool = NULL;
}


int FONTDESC::CreateIt(void)
{
	HDC hdc;

	if(hTool)
	{
		// Make sure a tool doesn't already exist...
		DeleteObject(hTool);
		hTool = NULL;
	}

	if(TypeCode == Font)
	{
		hdc = GetDC(0);

		// Create a font.
		LOGFONT lf;

		// Clear fields first.
		memset(&lf, 0, sizeof(lf));

		// Convert point size to MM_TEXT pixels.
		lf.lfHeight = -MulDiv(nPoints, GetDeviceCaps(hdc, LOGPIXELSY), 72);

		ReleaseDC(0, hdc);

		// If Bold flag is on, create a bold font--otherwise normal...
		lf.lfWeight = bIsBold ? FW_BOLD : FW_NORMAL;

		lf.lfOutPrecision = OUT_DEFAULT_PRECIS;
		lf.lfClipPrecision = CLIP_DEFAULT_PRECIS;
		lf.lfQuality = DEFAULT_QUALITY;
		lf.lfPitchAndFamily = DEFAULT_PITCH | FF_DONTCARE;
		lstrcpy(lf.lfFaceName, szFaceName);

		// Return created font
		hTool = CreateFontIndirect(&lf);
	}
	else
		MessageBox(NULL, "Warning: Unexpected Tool Type in TOOLDESC::CreateIt()", "Error", MB_ICONSTOP);
	if(hTool)
		return TRUE;
	else
		return FALSE;
}


SDrawCore::SDrawCore(HWND hwnd)
{
	bBufferScreenUpdates = FALSE;

	hdcActive = NULL;
	hdcDefault = NULL;
	hdcOffScreen = NULL;
	hdcMemory = NULL;

	nBitsPerPixelOffscreen = 0;

	cxMaxScreen = GetSystemMetrics(SM_CXFULLSCREEN);
	cyMaxScreen = GetSystemMetrics(SM_CYFULLSCREEN);

	hpenOld = 0;
	hbrOld = 0;
	hfontOld = 0;

	GDIToolsList = new LINKEDPTRLIST();
	if(!GDIToolsList)
	{
		MessageBox(NULL, "Can't create GDIToolList.", "Can't Draw", MB_ICONSTOP);
		return;
	}

	HDC hdcScreen = GetDC(hwnd);
	if(!hdcScreen)
	{
		delete GDIToolsList;
		GDIToolsList = NULL;
		MessageBox(NULL, "Can't get device context.", "Can't Draw", MB_ICONSTOP);
		return;
	}

	nBitsPerPixelOffscreen = GetDeviceCaps(hdcScreen, BITSPIXEL);

	hdcMemory = CreateCompatibleDC(hdcScreen);
	if(!hdcMemory)
	{
		delete GDIToolsList;
		GDIToolsList = NULL;
		ReleaseDC(hwnd, hdcScreen);
		MessageBox(NULL, "Can't create memory device context.", "Can't Draw", MB_ICONSTOP);
		return;
	}

	hdcOffScreen = CreateCompatibleDC(hdcScreen);
	if(!hdcOffScreen)
	{
		delete GDIToolsList;
		GDIToolsList = NULL;
		ReleaseDC(hwnd, hdcScreen);
		DeleteDC(hdcMemory);
		hdcMemory = NULL;
		MessageBox(NULL, "Can't create copy of device context.", "Can't Draw", MB_ICONSTOP);
		return;
	}

	hdcActive = FALSE;

	ReleaseDC(hwnd, hdcScreen);

	hbmCurrentOffscreen = CreateCompatibleBitmap(hdcScreen, cxMaxScreen, cyMaxScreen);
	if(!hbmCurrentOffscreen)
	{
		delete GDIToolsList;
		GDIToolsList = NULL;
		DeleteDC(hdcMemory);
		hdcMemory = NULL;
		DeleteDC(hdcOffScreen);
		hdcOffScreen = NULL;
		MessageBox(NULL, "Can't create off-screen buffer bitmap.", "Can't Draw", MB_ICONSTOP);
		return;
	}

	// Make offscreen buffer bitmap current.
	hbmDefaultOffscreenBitmap = SelectObject(hdcOffScreen, hbmCurrentOffscreen);
}


SDrawCore::~SDrawCore()
{
	// Clean-up off-screen buffers
	// TODO Add tool clean-up.
	TOOLDESC *ThisTool;

	if(hdcMemory)
		DeleteDC(hdcMemory);

	if(hbmDefaultOffscreenBitmap)
		SelectObject(hdcOffScreen, hbmDefaultOffscreenBitmap);
	if(hdcOffScreen)
		DeleteDC(hdcOffScreen);

	// Clean-up off-screen bitmap too...
	if(hbmCurrentOffscreen)
		DeleteObject(hbmCurrentOffscreen);

	// Clean-up resources.
	// (Assumes none of them are currently selected into
	// an active DC.)
	for(PTRNODE *Find = (PTRNODE *)GDIToolsList->Head->lpNext;
		Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		ThisTool = (TOOLDESC *)Find->lpData;
		if(ThisTool->hTool)
		{
			DeleteObject(ThisTool->hTool);
			ThisTool->hTool = NULL;
		}
	}

   // Clean-up list used to store GDI tools...
	if(GDIToolsList)
	{
		delete GDIToolsList;
		GDIToolsList = NULL;
	}
}


// Searches for existing Brush/Pen GDI object...
TOOLDESC *SDrawCore::DoesGDIToolExist(int Type, DWORD rgbColor, int nPenWidth)
{
	TOOLDESC *ThisTool;

	for(PTRNODE *Find = (PTRNODE *)GDIToolsList->Head->lpNext;
		Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		ThisTool = (TOOLDESC *)Find->lpData;
		if(ThisTool->hTool)
		{
			if(Type == Brush && ThisTool->TypeCode == Brush)
			{
				if(ThisTool->rgbColor == rgbColor)
					 return ThisTool;
			}
			else if(Type == Pen && ThisTool->TypeCode == Pen)
			{
				if(ThisTool->rgbColor == rgbColor && ThisTool->nPenWidth == nPenWidth)
					 return ThisTool;
			}
		}
	}
	return NULL;
}


TOOLDESC *SDrawCore::AddGDITool(int Type, DWORD rgbColor, int nPenWidth)
{
	int rc;

	if(Type != Pen && Type != Brush)
	{
		MessageBox(NULL, "Wrong type in SDrawCore::AddGDITool()", "Error", MB_ICONSTOP);
		return NULL;
	}

	TOOLDESC *ThisTool = new TOOLDESC((eToolType)Type, rgbColor, nPenWidth);
	if(!ThisTool)
	{
		MessageBox(NULL, "Can't allocate memory in SDrawCore::AddGDITool()", "Error", MB_ICONSTOP);
		return NULL;
	}

	// Try to create our GDI tool.
	rc = ThisTool->CreateIt();
	if(!rc)
	{
		MessageBox(NULL, "Can't create new drawing tool in SDrawCore::AddGDITool()", "Error", MB_ICONSTOP);
		delete ThisTool;
		return NULL;
	}

	// Add to our linked list.
	rc = GDIToolsList->AddPreexistingNode((LPSTR)ThisTool);
	if(!rc)
	{
		MessageBox(NULL, "Can't store new drawing tool in SDrawCore::AddGDITool()", "Error", MB_ICONSTOP);
		DeleteObject(ThisTool->hTool);
		delete ThisTool;
		return NULL;
	}
	return ThisTool;
}


TOOLDESC *SDrawCore::GetTool(eToolType Type, DWORD rgbColor, int nPenWidth)
{
	TOOLDESC *ThisTool = DoesGDIToolExist(Type, rgbColor, nPenWidth);

	// If this tool doesn't exist, create it.
	if(!ThisTool)
		ThisTool = AddGDITool(Type, rgbColor, nPenWidth);

	return ThisTool;
}


// Searches for existing Font GDI object...
FONTDESC *SDrawCore::DoesFontExist(char *lpszFaceName, int nPoints, int bIsBold)
{
	TOOLDESC *ThisTool;
	FONTDESC *ThisFont;

	for(PTRNODE *Find = (PTRNODE *)GDIToolsList->Head->lpNext;
		Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		ThisTool = (TOOLDESC *)Find->lpData;
		if(ThisTool->TypeCode != Font)
			continue;
		ThisFont = (FONTDESC *)Find->lpData;
		if(ThisFont->hTool && strcmpi(ThisFont->szFaceName, lpszFaceName) == 0 &&
			ThisFont->nPoints == nPoints && ThisFont->bIsBold == bIsBold)
			return ThisFont;
	}
	return NULL;
}


FONTDESC *SDrawCore::AddFont(char *lpszFaceName, int nPoints, int bIsBold)
{
	int rc;

	FONTDESC *ThisFont = new FONTDESC(lpszFaceName, nPoints, bIsBold);

	if(!ThisFont)
	{
		MessageBox(NULL, "Can't allocate memory in SDrawCore::AddFont()", "Error", MB_ICONSTOP);
		return NULL;
	}

	// Try to create our GDI font.
	rc = ThisFont->CreateIt();
	if(!rc)
	{
		MessageBox(NULL, "Can't create new font in SDrawCore::AddFont()", "Error", MB_ICONSTOP);
		delete ThisFont;
		return NULL;
	}

	// Add to our linked list.
	rc = GDIToolsList->AddPreexistingNode((LPSTR)ThisFont);
	if(!rc)
	{
		MessageBox(NULL, "Can't store new drawing tool in SDrawCore::AddGDITool()", "Error", MB_ICONSTOP);
		DeleteObject(ThisFont->hTool);
		delete ThisFont;
		return NULL;
	}
	return ThisFont;
}


FONTDESC *SDrawCore::GetFont(LPSTR lpszFaceName, int nPoints, int bIsBold)
{
	FONTDESC *ThisFont = DoesFontExist(lpszFaceName, nPoints, bIsBold);

	// If this tool doesn't exist, create it.
	if(!ThisFont)
		ThisFont = AddFont(lpszFaceName, nPoints, bIsBold);
	else
		ThisFont;

	return ThisFont;
}


BOOL SDrawCore::SelectBrush(HDC hdc, DWORD rgbColor)
{
	TOOLDESC *ThisTool;

	if(rgbColor == Transparent)
		hbrOld = SelectObject(hdc, GetStockObject(NULL_BRUSH));
	else
		ThisTool = GetTool(Brush, rgbColor);

	if(ThisTool)
	{
		hbrOld = SelectObject(hdc, ThisTool->hTool);
		return TRUE;
	}
	else
		return FALSE;
}


BOOL SDrawCore::SelectPen(HDC hdc, DWORD rgbColor, int nPenWidth)
{
	TOOLDESC *ThisTool;

	ThisTool = GetTool(Pen, rgbColor, nPenWidth);

	if(ThisTool)
	{
		hpenOld = SelectObject(hdc, ThisTool->hTool);
		return TRUE;
	}
	else
		return FALSE;
}


BOOL SDrawCore::SelectFont(HDC hdc, char *lpszFaceName, int nPoints, int bIsBold)
{
	FONTDESC *ThisFont;

	ThisFont = GetFont(lpszFaceName, nPoints, bIsBold);

	if(ThisFont)
	{
		hfontOld = SelectObject(hdc, ThisFont->hTool);
		return TRUE;
	}
	else
		return FALSE;
}


void SDrawCore::RestoreOldBrush(HDC hdc)
{
	SelectObject(hdc, hbrOld);
}


void SDrawCore::RestoreOldPen(HDC hdc)
{
	SelectObject(hdc, hpenOld);
}


void SDrawCore::RestoreOldFont(HDC hdc)
{
	SelectObject(hdc, hfontOld);
}


void SDrawCore::SetDrawingSurface(HDC hdcNew)
{
	if(hdcNew == 0)
		hdcDefault = GetDC(0);
	else
		hdcDefault = hdcNew;
}


void SDrawCore::CopyScreenToBuffer(HWND hwnd, int cxCopy, int cyCopy)
{
	hdcDefault = GetDC(hwnd);

	if(cxCopy && cyCopy)
		// Copy currently displayed area of screen.
		BitBlt(hdcOffScreen, 0, 0, cxCopy, cyCopy,
				 hdcDefault, 0, 0, SRCCOPY);
	else
		// Copy whole screen by default.
		BitBlt(hdcOffScreen, 0, 0, cxMaxScreen, cyMaxScreen,
			 hdcDefault, 0, 0, SRCCOPY);

	ReleaseDC(hwnd, hdcDefault);
	hdcDefault = 0;
}


void SDrawCore::CopyBufferToScreen(HWND hwnd, int cxCopy, int cyCopy)
{
	hdcDefault = GetDC(hwnd);

	if(cxCopy && cyCopy)
		// Copy currently displayed area of screen.
		BitBlt(hdcDefault, 0, 0, cxCopy, cyCopy,
				 hdcOffScreen, 0, 0, SRCCOPY);
	else
		// Copy whole screen by default.
		BitBlt(hdcDefault, 0, 0, cxMaxScreen, cyMaxScreen,
				 hdcOffScreen, 0, 0, SRCCOPY);

	ReleaseDC(hwnd, hdcDefault);
	hdcDefault = 0;
}


void SDrawCore::StartScreenBuffering(void)
{
	hdcActive = hdcOffScreen;
	bBufferScreenUpdates = TRUE;
}


void SDrawCore::StopScreenBuffering(void)  //I.e. write directly to screen!
{
	hdcActive = hdcDefault;
	bBufferScreenUpdates = FALSE;
}


void SDrawCore::ClearOffscreenBuffer(void)
{
	PatBlt(hdcOffScreen, 0, 0, cxMaxScreen, cxMaxScreen, BLACKNESS);
}

#if defined (INCL_DEBUG)
void SDrawCore::ReportToolUsage(HWND hwnd)
{
	TOOLDESC *ThisTool;

	char szMsg[256];

	// Reports tool usage to Window's caption.
	int nBrushCount = 0;
	int nBadBrushCount = 0;

	int nPenCount = 0;
	int nBadPenCount = 0;

	int nFontCount = 0;
	int nBadFontCount = 0;

	for(PTRNODE *Find = (PTRNODE *)GDIToolsList->Head->lpNext;
		Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		ThisTool = (TOOLDESC *)Find->lpData;
		if(ThisTool->TypeCode != Font)
		{
			if(ThisTool->TypeCode == Brush)
			{
				nBrushCount++;
				if(ThisTool->hTool == 0)
					nBadBrushCount++;
			}
			else if(ThisTool->TypeCode == Pen)
			{
				nPenCount++;
				if(ThisTool->hTool == 0)
					nBadPenCount++;
			}
			else
				MessageBeep(0);
		}
		else
		{
			// We're a font...
			nFontCount++;
			if(ThisTool->hTool == 0)
				nBadFontCount++;
		}
	}

	// Construct debug info.
	wsprintf(szMsg, "Bru=%d(%d bad); Pen=%d(%d bad); Fnt=%d(%d bad)",
				nBrushCount, nBadBrushCount, nPenCount, nBadPenCount,
				nFontCount, nBadFontCount);
	SetWindowText(hwnd, szMsg);
}
#endif
