/*********************************************************
	SWINDOWS.CPP -- Implementation file of "simple" Windows
						 class library.

					  Designed to allow (beginning) non-Windows
					  C++ programmers to get at Windows
					  functionality without all the API overhead.

					  Current version encapsulates:

							-- 16-bit and 32-bit Windows support.
							-- Basic GDI drawing
								(including bitmap rendering).
							--	Off-screen buffering for
								non-destructive drawing.
							-- Painting.
                     --	Servicing and modifying
								menu items.
							-- Idle-time processing (for
								simulations and screensavers).
							-- Timer messages (for background
								processing, e.g. simple games).
							--	Servicing and modifying menu items
								and run-time.
							-- Keystroke mapping to specific
								methods.
							-- Mouse events (Left and right
								mouse buttons).
							-- A customized 'About' dialog box.
							-- The ability to create custom
								modal dialog boxes (This requires
								more Windows API knowledge).

							"What's the Code?"
							Computer Shopper -- May 1995

							v. 1.00 -- by Richard Dragan

******************************************************/
#include "swindows.h"
#include "sdialog.h"

#include "resids.h"

#if defined (WIN32)
inline void MAKEPOINT32(POINT *lpPoint, LPARAM lParam)
{
	lpPoint->x = LOWORD(lParam);
	lpPoint->y = HIWORD(lParam);
}
#endif

// Globals
char szDefaultAppName[] = "SWindows Application";

SWindows *pActiveObject = NULL;

// Dummy proc. used for our main window
// (We'll be dispatching messages by ourselves.)
LRESULT FAR PASCAL _export DummyWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	return pActiveObject->MainWndProc(hwnd, msg, wParam, lParam);
}

// Constructor
SWindows::SWindows()
{
	// Initialize all our members.
	nErrCode = 0;
	_fmemset(szErrMsg, 0, 80);

	pGraphics = NULL;

	pActiveObject = this;

	hwnd = 0;
	hmenu = 0;
	hIconApp = 0;
	hInst = 0;
	hPrevInst = 0;

	lstrcpy(szCaption, szDefaultAppName);
	_fmemset(szCommandLine, 0, 128);
	nShowCommand = SW_SHOWNORMAL;

	X = CW_USEDEFAULT;
	Y = CW_USEDEFAULT;
	CX = CW_USEDEFAULT;
	CY = CW_USEDEFAULT;

	bMessageHandled = FALSE;

	SetAboutBoxPrompts("SWindows Sample Application",
							 "        Version 1.00       ",
							 NULL, "About");

	// REM SetAboutBoxIcon(szDefaultAppName);

	lstrcpy(szDefMenuName, "MAIN_MENU");
	bMenuNeedsDestroying = FALSE;

	// Create linked lists.
	MenuItemHandlerList = new LINKEDPTRLIST();
	if(!MenuItemHandlerList)
	{
		SetError(ALLOC_FAILED, "Can't create Menu Item linked list.");
		return;
	}

	TimerList = new LINKEDPTRLIST();
	if(!TimerList)
	{
		SetError(ALLOC_FAILED, "Can't create Menu Item linked list.");
		return;
	}

	IdleEventHandler = FALSE;

	//REM SetAboutBoxMenuItem(IDM_ABOUT, (HANDLERPROC)::CmHelpAbout);
	EnableAboutBox();
}

// Destructor
SWindows::~SWindows()
{
	// Clean-up.
	if(MenuItemHandlerList)
	{
		delete MenuItemHandlerList;
		MenuItemHandlerList = NULL;
	}

	if(TimerList)
	{
		delete TimerList;
		TimerList = NULL;
	}

	if(hIconApp)
	{
		DeleteObject(hIconApp);
		hIconApp = NULL;
	}

	if(bMenuNeedsDestroying && hmenu)
	{
		DestroyMenu(hmenu);
		hmenu = NULL;
	}
}


int SWindows::InitApp(HINSTANCE hInst,
										HINSTANCE hPrevInst,
										LPSTR lpszCmdLine,
										int nCmdShow,
										UINT uWinStyle)
{
	WNDCLASS wndclass;

	UINT uUseWinStyle = uWinStyle ? uWinStyle : CS_HREDRAW | CS_VREDRAW;

	SetHInstance(hInst);
	SetHPrevInstance(hPrevInst);
	SetCommandLine(lpszCmdLine);
	SetShowWindowCommand(nCmdShow);

	// Registers our windows class.
	if (!hPrevInst)
	{
		wndclass.style          =  uUseWinStyle;
		wndclass.lpfnWndProc    =  DummyWndProc;
		wndclass.cbClsExtra     =  0 ;
		wndclass.cbWndExtra     =  0 ;
		wndclass.hInstance      =  hInst ;
		wndclass.hIcon          =  LoadIcon (hInst, "SWindows");
		wndclass.hCursor        =  LoadCursor (NULL, IDC_ARROW);
		wndclass.hbrBackground  =  GetStockObject(WHITE_BRUSH);
		wndclass.lpszMenuName   =  szDefMenuName;
		wndclass.lpszClassName  =  "SWindowsClass" ;

		if (!RegisterClass (&wndclass))
			return FALSE ;
		else
			return TRUE;
	}
	else
		return TRUE;
}


int SWindows::InitMainWindow(DWORD dwWinStyle)
{
	DWORD dwUseWinStyle = dwWinStyle ? dwWinStyle :
								 WS_OVERLAPPEDWINDOW;

	hwnd = CreateWindow ("SWindowsClass",
								szCaption,
								dwUseWinStyle,
								X, Y,
								CX, CY,
								NULL,
								NULL,
								hInst,
								NULL) ;

	if(!hwnd)
		return FALSE;
	else
	{
		hmenu = GetMenu(hwnd);
		ShowWindow (hwnd, nShowCommand);   // Make window visible
		UpdateWindow (hwnd) ;              // Update window
	}

	return TRUE;
}


void SWindows::DoMessageLoop(void)
{
	MSG msg;

	while(TRUE)
	{
		if(PeekMessage(&msg, NULL, 0, 0, PM_REMOVE)) // Message queue is NOT empty;
																	// deliver message just like GetMessage()
		{
			if(msg.message == WM_QUIT) // Exit loop on WM_QUIT message
				break;
			else
			{
				TranslateMessage(&msg);
				DispatchMessage(&msg);
			}
		}
		else // Message queue is empty; call IdleTimeProc();
		{
			if(IdleEventHandler)
				(*IdleEventHandler)(this);
			else
			{
				if(pGraphics)
					OnIdleEvent(hwnd, *pGraphics); 
#if !defined(MSC_VER)					
				else
					OnIdleEvent(hwnd, *pGraphics); 
#endif					
			}
		}
	}
}


LRESULT SWindows::MainWndProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
	POINT ptMouse;

	hwnd;  // Kill warning.

	bMessageHandled = FALSE;

	switch(msg)
	{
		case WM_CREATE: // Init. message
			{
				if(CX == CW_USEDEFAULT || CY == CW_USEDEFAULT)
				{
					// Get window size in case we're using default size
					RECT rClient;
					GetClientRect(hwnd, &rClient);
					CX = rClient.right;
					CY = rClient.bottom;
				}
				OnCreate(hwnd, lParam);
			}
			return 0L;

		case WM_SIZE:   // Size message --
			// Watch out for 32-bits here
			{
				int cxSize = LOWORD(lParam);
				int cySize = HIWORD(lParam);

				CX = cxSize;
				CY = cySize;

				OnSize(cxSize, cySize);
			}
			return 0L;

		case WM_COMMAND: // Look at commands
			{
#ifdef WIN32
				WORD wId = LOWORD(wParam);
#else
				WORD wId = wParam;
#endif
				OnCommand(wId);
			}
			return 0L;

		case WM_PAINT:  // Paint our window.
			OnPaint(*pGraphics);
			if(bMessageHandled)
				return 0L;
			else
				return DefWindowProc(hwnd, msg, wParam, lParam);

		case WM_KEYDOWN:  // Key press -- VK_CODE
			OnEvKey(wParam);
			return 0L;

		case WM_CHAR:     // Key press ANSI/ASCII char.
			OnEvChar(wParam);
			return 0L;

		case WM_TIMER:    // Timer event.  Pass along ID.
			OnTimer(wParam);
			return 0L;

		// Mouse messages--these require separate 32-bit processing
		case WM_LBUTTONDOWN:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnLButtonDown(ptMouse);
			return 0L;

		case WM_LBUTTONUP:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnLButtonUp(ptMouse);
			return 0L;

		case WM_LBUTTONDBLCLK:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnLButtonDblClick(ptMouse);
			return 0L;

		case WM_RBUTTONDOWN:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnRButtonDown(ptMouse);
			return 0L;

		case WM_RBUTTONUP:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnRButtonUp(ptMouse);
			return 0L;

		case WM_RBUTTONDBLCLK:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnRButtonDblClick(ptMouse);
			return 0L;

		case WM_MOUSEMOVE:
#if defined (WIN32)
			MAKEPOINT32(&ptMouse, lParam);
#else
			ptMouse = MAKEPOINT(lParam);
#endif
			OnMouseMove(ptMouse);
			return 0L;

		case WM_DESTROY: // Clean-up
			OnDestroy();
			PostQuitMessage(0);
			return 0L;

		default: return DefWindowProc(hwnd, msg, wParam, lParam);
	}
}


BOOL SWindows::InitGraphics(HWND hwnd, int bUseOffscreenBuffering)
{
	pGraphics = new SDraw(hwnd);
	if(!pGraphics)
	{
		SetError(SCREENBUFFER_INIT_FAILED, "Can't initialize graphics module.");
		return FALSE;
	}

	if(bUseOffscreenBuffering)
		pGraphics->StartScreenBuffering();

	return TRUE;
}


void SWindows::CleanupGraphics(void)
{
	if(!pGraphics)
	{
		delete pGraphics;
		pGraphics = NULL;
	}
}


void SWindows::OnCreate(HWND hwnd, LPARAM lParam)
{
	// Do SScreen object core initialization.
	InitGraphics(hwnd, FALSE);
	AddMenuItemEventHandler(IDM_EXIT, (HANDLERPROC)CmFileExit); 
	AddMenuItemEventHandler(IDM_ABOUT, (HANDLERPROC)CmHelpAbout);

	bMessageHandled = TRUE;  // Nothing else to do.

	lParam;  // Kill warning.
}


void SWindows::OnSize(int xSize, int ySize)
{
	CX = xSize;
	CY = ySize;
	bMessageHandled = TRUE;
}


void SWindows::OnCommand(int nCommand)
{
	MENUITEMHANDLERREC *lpThisRec;
	HANDLERPROC lpfnMenuHandler = NULL;

	// Walk through Menu Item handlers
	// and see if we have a match.

	// If so, call appropriate handler.
	for(PTRNODE *Find = (PTRNODE *)MenuItemHandlerList->Head->lpNext;
		 Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		lpThisRec = (MENUITEMHANDLERREC *)Find->lpData;
		if(lpThisRec->nMenuItemId == nCommand)
		{
			lpfnMenuHandler = lpThisRec->lpfnHandler;
			break;
		}
	}

	if(lpfnMenuHandler)
	{
		(*lpfnMenuHandler)(this);
	}
	bMessageHandled = TRUE;
}


void SWindows::OnDestroy()
{
	StopAllTimers();
	if(pGraphics)
	{
		delete pGraphics;
		pGraphics = NULL;
	}

	bMessageHandled = TRUE;
}


void CmFileExit(SWindows *pObj)
{
	pObj->CloseApp();
}

void CmHelpAbout(SWindows *pObj)
{
	// Provides default processing for an 'About Dialog'.
	SAboutDlg DefAboutDlg(pObj->hInst, pObj->hwnd, (LPSTR)"ABOUT_DLG");

	DefAboutDlg.SetAboutPrompts(pObj->szAboutBoxLine1Text,
										 pObj->szAboutBoxLine2Text,
										 pObj->szAboutBoxLine3Text,
										 pObj->szAboutBoxCaption);
	DefAboutDlg.Execute();
}


void SWindows::SetCaption(LPSTR lpszCaption)
{
	if(!lpszCaption)
	{
		SetWindowText(hwnd, "");
		_fmemset(szCaption, 0, 80);
	}
	else
	{
		SetWindowText(hwnd, lpszCaption);
		lstrcpy(szCaption, lpszCaption);
	}
}


void SWindows::SetIcon(HICON hIcon)
{
#if defined (WIN32)
	SetClassLong(hwnd, GCL_HICON, (LONG)hIcon);
#else
	SetClassWord(hwnd, GCW_HICON, hIcon);
#endif
	if(hIconApp)
		DeleteObject(hIconApp);
	hIconApp = hIcon;
}


void SWindows::SetIcon(LPSTR lpszIconResourceName)
{
	HICON hIconNew = LoadIcon(hInst, lpszIconResourceName);

	if(hIconNew)
		SetIcon(hIconNew);
}


void SWindows::ShowWaitCursor(void)
{
	SetCursor(LoadCursor(NULL, IDC_WAIT));
}


void SWindows::ShowArrowCursor(void)
{
	SetCursor(LoadCursor(NULL, IDC_ARROW));
}


void SWindows::RepaintWindow(void)
{
	InvalidateRect(hwnd, 0, TRUE);
	UpdateWindow(hwnd);
}


void SWindows::MoveWindow(int nX, int nY, int nCX, int nCY, int bRedraw)
{
	::MoveWindow(hwnd, nX, nY, nCX, nCY, bRedraw);
	UpdateWindow(hwnd);
}


void SWindows::CloseApp(void)
{
	DestroyWindow(hwnd);
}


int SWindows::MsgBox(LPSTR lpszText, LPSTR lpszCaption, int nOptions)
{
	// Pass back user's choice where appropriate.
	return ::MessageBox(hwnd, lpszText, lpszCaption, nOptions);
}

#if defined (INCL_DEBUG)
void SWindows::Dbg(LPSTR lpszStringVal, LPSTR lpszLabel)
{
	char szTemp[256];
	wsprintf(szTemp, "%-s='%-s'", lpszLabel, lpszStringVal);
	MsgBox(szTemp, "Debugging", MB_ICONINFORMATION);
}

void SWindows::Dbg(int nVal, LPSTR lpszLabel)
{
	char szTemp[256];
	wsprintf(szTemp, "%-s=%d", lpszLabel, nVal);
	MsgBox(szTemp, "Debugging", MB_ICONINFORMATION);
}

void SWindows::Dbg(UINT uVal, LPSTR lpszLabel)
{
	char szTemp[256];
	wsprintf(szTemp, "%-s=%u", lpszLabel, uVal);
	MsgBox(szTemp, "Debugging", MB_ICONINFORMATION);
}

void SWindows::Dbg(LONG lVal, LPSTR lpszLabel)
{
	char szTemp[256];
	wsprintf(szTemp, "%-s=%ld", lpszLabel, lVal);
	MsgBox(szTemp, "Debugging", MB_ICONINFORMATION);
}

void SWindows::Dbg(DWORD dwVal, LPSTR lpszLabel)
{
	char szTemp[256];
	wsprintf(szTemp, "%-s=%lu", lpszLabel, dwVal);
	MsgBox(szTemp, "Debugging", MB_ICONINFORMATION);
}
#endif


void SWindows::SetMenu(LPSTR lpszMenuName)
{
	HMENU hmenuNew = LoadMenu(hInst, lpszMenuName);
	if(!hmenuNew)
	{
		char szMsg[80];
		wsprintf(szMsg,"Can't load menu resource '%-s'", lpszMenuName);
		SetError(RESOURCE_LOAD_FAILED, szMsg);
		return;
	}

	HMENU hmenuOld = GetMenu(hwnd);

	::SetMenu(hwnd, hmenuNew);
	bMenuNeedsDestroying = TRUE;
	if(hmenuOld)
	{
		DestroyMenu(hmenuOld);
	}

	hmenu = hmenuNew;
}


int SWindows::AddMenuItemEventHandler(int nMenuItemId, HANDLERPROC lpfnHandler)
{
	int rc;
	int bRecFound = FALSE;

	MENUITEMHANDLERREC NewRec;
	MENUITEMHANDLERREC *ThisRec;

	NewRec.nMenuItemId = nMenuItemId;
	NewRec.lpfnHandler = lpfnHandler;

	// Check to make sure new ID isn't already in list.
	for(PTRNODE *Find = (PTRNODE *)MenuItemHandlerList->Head->lpNext;
		 Find != NULL; Find=(PTRNODE *)Find->lpNext)
	{
		ThisRec = (MENUITEMHANDLERREC *)Find->lpData;
		if(ThisRec->nMenuItemId == nMenuItemId)
		{
			_fmemcpy(ThisRec, &NewRec, sizeof(NewRec));
			bRecFound = TRUE;
			break;
		}
	}

	if(!bRecFound)
	{
		rc = MenuItemHandlerList->AddNode((LPSTR)&NewRec, sizeof(NewRec));

		if(!rc)
			return FALSE;
		else
			return TRUE;
	}
	return TRUE;  // Kill warning.
}


void SWindows::ClearMenuItemEventHandlers(void)
{
	MenuItemHandlerList->EmptyList();
}


void SWindows::CheckMenuItem(int nMenuItemId)
{
	::CheckMenuItem(hmenu, nMenuItemId, MF_CHECKED);
}


void SWindows::CheckMenuItem(int nMenuItemId, int bIsChecked)
{
	::CheckMenuItem(hmenu, nMenuItemId, bIsChecked ? MF_CHECKED : MF_UNCHECKED);
}


void SWindows::UncheckMenuItem(int nMenuItemId)
{
	::CheckMenuItem(hmenu, nMenuItemId, MF_UNCHECKED);
}


void SWindows::EnableMenuItem(int nMenuItemId)
{
	::EnableMenuItem(hmenu, nMenuItemId, MF_ENABLED);
}


void SWindows::DisableMenuItem(int nMenuItemId)
{
	::EnableMenuItem(hmenu, nMenuItemId, MF_GRAYED);
}


void SWindows::EnableMenuItem(int nMenuItemId, int bEnabled)  // (Emulates API call.)
{
	::EnableMenuItem(hmenu, nMenuItemId, bEnabled ? MF_ENABLED : MF_GRAYED);
}


void SWindows::SetMenuItemPrompt(int nMenuItemId, LPSTR lpszMenuText)
{
	// Changes string to show in menu.
	::ModifyMenu(hmenu, nMenuItemId, MF_BYCOMMAND | MF_STRING, nMenuItemId, lpszMenuText);
}


int SWindows::StartTimer(int nTimerId, UINT uMilliSeconds, HANDLERPROC lpfnTimerProc)
{
	int rc;
	rc = ::SetTimer(hwnd, nTimerId, uMilliSeconds, (FARPROC)lpfnTimerProc);
	if(rc)
		AddTimerToList(nTimerId);
	return rc;
}


int SWindows::StopTimer(int nTimerId)
{
	return ::KillTimer(hwnd, nTimerId);
}


void SWindows::StopAllTimers(void)
{
	for(PTRNODE *Find = (PTRNODE *)TimerList->Head->lpNext;
		 Find != NULL; Find = (PTRNODE *)Find->lpNext)
	{
		KillTimer(hwnd, *((TIMERID *)Find->lpData));
	}

	TimerList->EmptyList();
}


int SWindows::AddTimerToList(int nTimerId)
{
	int rc;

	rc = TimerList->AddNode((LPSTR)&nTimerId, sizeof(int));

	if(!rc)
		return FALSE;
	else
		return TRUE;
}


void SWindows::SetIdleTimeEventHandler(HANDLERPROC lpfnHandler)
{
	IdleEventHandler = lpfnHandler;
}


void SWindows::ClearIdleTimeEventHandler(void)
{
	IdleEventHandler = NULL;
}


void SWindows::SetAboutBoxMenuItem(int nMenuId, HANDLERPROC lpfnAboutHandler)
{
	AddMenuItemEventHandler(nMenuId, lpfnAboutHandler);
}


void SWindows::EnableAboutBox()
{
	bEnableAboutBox = TRUE;
}


void SWindows::DisableAboutBox()
{
	bEnableAboutBox = FALSE;
}


void SWindows::SetAboutBoxPrompts(LPSTR lpszLine1, LPSTR lpszLine2, LPSTR lpszLine3, LPSTR lpszCaption)
{
	if(lpszLine1)
		lstrcpy(szAboutBoxLine1Text, lpszLine1);
	else
		_fmemset(szAboutBoxLine1Text, 0, 80);

	if(lpszLine2)
		lstrcpy(szAboutBoxLine2Text, lpszLine2);
	else
		_fmemset(szAboutBoxLine2Text, 0, 80);

	if(lpszLine3)
		lstrcpy(szAboutBoxLine3Text, lpszLine3);
	else
		_fmemset(szAboutBoxLine3Text, 0, 80);

	if(lpszCaption)
		lstrcpy(szAboutBoxCaption, lpszCaption);
	else
		_fmemset(szAboutBoxCaption, 0, 80);
}


// Error routines -- Structured exception handling
//                   will be added in a later version.
int SWindows::GetLastError(LPSTR lpszErrDesc)
{
	if(lpszErrDesc)
		lstrcpy(lpszErrDesc, szErrMsg);
	return nErrCode;
}


void SWindows::ClearError(void)
{
	nErrCode = 0;
	_fmemset(szErrMsg, 0, 80);
}


void SWindows::SetError(int nErr, LPSTR lpszErrMsg)
{
	nErrCode = nErr;
	if(lpszErrMsg)
		lstrcpy(szErrMsg, lpszErrMsg);
}

// Globals needs to pass info. from WinMain():
HINSTANCE hGlobalInst;
HINSTANCE hGlobalPrevInst;
char achGlobalCmdLine[128];
int  nGlobalCmdShow;


void SWindows::RunIt(void)
{
	InitApp(hGlobalInst, hGlobalPrevInst, achGlobalCmdLine, nGlobalCmdShow, NULL);
	InitMainWindow();
	DoMessageLoop();
}


// Entry point into a Windows application.
int PASCAL WinMain (HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
	// Save globals.
	hGlobalInst = hInstance;
	hGlobalPrevInst = hPrevInstance;
	lstrcpy(achGlobalCmdLine, lpCmdLine);
	nGlobalCmdShow = nCmdShow;

	// Call C main() function;
	Main();

	nCmdShow;  // Kill warning.

	return 0;
}

