/*********************************************************
	SWIN_SIM.CPP -- Custom code for an SWindows application.
						 (Demonstrates simulation.)

					 Directions for use:

					 1) Derive your own class based on
						 SWindows base class.

					 2) Customize those methods that
						 you need (for example OnPaint())
						 in your derived class.

					 3) In your main() routine,
						 create an object of your derived class.

					 4) Run it with the SWindows RunIt() method.

**********************************************************/
#include "swindows.h"
#include "sim.h"
#include <stdlib.h>

#include "resids.h"

int bInAnimation = FALSE;  // A global to track whether we're animating or not.

void Main(void);

class MySimulationApp : public SWindows
{
public:
	// Constructor -- just call base class SWindows constructor
	MySimulationApp() : SWindows() { SetCaption("SWindows (Simple Windows C++ Library)--Simulation with Sprites");
												bInAnimation = FALSE;
												GatorSprite = NULL; }

	// Destructor.
	~MySimulationApp() { delete GatorSprite; }

	// Menu item handlers
	// Note the form here:

	// friend void CommandHandlerName(SWindows *pObj);

	// These must be C friend functions.
	// Also, you'll need to add a menu item handler in your
	// OnCreate() method. This attaches the C function to a particular menu event.

	friend void CmStartStopAnimation(SWindows *pObj);

	virtual void OnCreate(HWND hwnd, LPARAM lParam);

	virtual void OnSize(int cxSize, int cySize);

	virtual void OnPaint(SDraw& Graphics);  // Custom painting

	virtual void OnIdleEvent(HWND hwnd, SDraw& Graphics);  // Do Animation.

	virtual void OnDestroy(void);

	// Simulation Objects
	SimCentipede *Centipede;

	SimMushroom *Mushrooms;

	SimAlligator *Gator;

	SSprite *GatorSprite;  // Used to display the current Gator object onscreen.
};


void CmStartStopAnimation(SWindows *pObj)
{
	if(bInAnimation == TRUE)
	{
		// Then pause it and change menu prompt.
		bInAnimation = FALSE;
		pObj->SetMenuItemPrompt(IDM_START_STOP_ANIM, "&Resume Animation");
	}
	else
	{
		// Then re-start the animation and change menu prompt.
		bInAnimation = TRUE;
		pObj->SetMenuItemPrompt(IDM_START_STOP_ANIM, "&Pause Animation");
	}
}


void MySimulationApp::OnCreate(HWND hwnd, LPARAM lParam)
{
	// Initialize graphics module -- 'TRUE' indicates use of off-screen buffering.
	InitGraphics(hwnd, TRUE);

	AddMenuItemEventHandler(IDM_EXIT, (HANDLERPROC)CmFileExit);   //&SWindows::CmFileExit
	AddMenuItemEventHandler(IDM_ABOUT, (HANDLERPROC)CmHelpAbout);
	AddMenuItemEventHandler(IDM_START_STOP_ANIM, (HANDLERPROC)CmStartStopAnimation);

	// Create a new centipede.
	Centipede = new SimCentipede(hwnd, "BMP_CENTIPEDE", "BMP_CENTIPEDE_MASK");
	Centipede->InitNewCentipede(-21, 12);
	Centipede->UpdateSegmentArray();

	// Create some mushrooms.
	Mushrooms = new SimMushroom[10];

	int yWin = 10;

	for(int i = 0; i < 10; i++)
	{
		if(i == 0)
			// First time through, load bitmaps--otherwise, just copy bitmap handles.
			Mushrooms[i].LoadBitmaps(hwnd, "BMP_MUSHROOM", "BMP_MUSHROOM_MASK");
		else
			Mushrooms[i] = Mushrooms[0];
		Mushrooms[i].InitPos(random(CX), yWin);
		yWin += Mushrooms[i].cy + 10;
	}

	// Create an alligator.
	Gator = new SimAlligator(hwnd);
	Gator->Init(hwnd);

	bMessageHandled = TRUE;

	bInAnimation = TRUE;
	lParam;  // Kill warning.
}


void MySimulationApp::OnSize(int cxSize, int cySize)
{
	// Update sprite dimensions.
	Centipede->SaveWindowSize(hwnd);

	if(Gator)
		Gator->SetCurrentWindowDim(hwnd);

	cxSize;
	cySize;  // Kill warnings.
}


// Do Animation.
void MySimulationApp::OnIdleEvent(HWND hwnd, SDraw& Graphics)
{
	int rc = TRUE;
	POINT pt;
	RECT rXOR, rAND;
	int bFirst = TRUE;

	if(!bInAnimation)
		return;

	Graphics.StartDrawing();

	// Erase the offscreen buffer first.
	Graphics.Rectangle(0, 0, CX, CY, Black);

	// Display each Centipede 'segment'
	while(rc)
	{
		if(bFirst)
		{
			rc = Centipede->GetFirstCentipedeSpriteSegment(pt, &rXOR, &rAND);
			bFirst = FALSE;
		}
		else
			rc = Centipede->GetNextCentipedeSpriteSegment(pt, &rXOR, &rAND);
		Graphics.DrawSprite(pt, *Centipede, rXOR, rAND);
	}

	// Calc. new position for Centipede.
	Centipede->CalcNewPos();
	Centipede->UpdateSegmentArray();

	// Draw Alligator and move for next time.
	Graphics.DrawCartoon(*Gator);
	Gator->UpdateNewPosition(TRUE);

	// Move and draw Mushrooms.
	for(int i = 0; i < 10; i++)
	{
		Mushrooms[i].UpdatePos();
		rc = Mushrooms[i].IsOnEdge(hwnd);
		if(rc)
			Mushrooms[i].ChangeDirection((eEdges)rc);
		Graphics.DrawSprite(Mushrooms[i]);
	}

	Graphics.StopDrawing();

	// Render screenful of data from buffer to screen.
	Graphics.CopyBufferToScreen(CX, CY);
}


void MySimulationApp::OnPaint(SDraw& Graphics)
{
	// Custom painting -- Here just erase screen.
	Graphics.StartPainting();

	Graphics.Rectangle(0, 0, CX, CY, Black);

	Graphics.StopPainting();
	bMessageHandled = FALSE;
}


void MySimulationApp::OnDestroy(void)
{
	// Clean-up simulation.
	delete[] Mushrooms;

	delete Centipede;

	delete Gator;

	CleanupGraphics();
}


/*********************************************************
	Main() -- This procedure is called first by SWindows.
*********************************************************/
void Main(void)
{
	// Create an (automatic) object based on the SWindows class.
	MySimulationApp MySimApp;

	// Run it!
	MySimApp.RunIt();
}
